package api

import (
	"encoding/json"
	"errors"
	"fmt"
	"strconv"

	"srs.epita.fr/fic-server/admin/sync"
	"srs.epita.fr/fic-server/libfic"

	"github.com/julienschmidt/httprouter"
)

func init() {
	router.GET("/api/themes", apiHandler(listThemes))
	router.POST("/api/themes", apiHandler(createTheme))
	router.GET("/api/themes.json", apiHandler(exportThemes))
	router.GET("/api/files-bindings", apiHandler(bindingFiles))

	router.GET("/api/themes/:thid", apiHandler(themeHandler(showTheme)))
	router.PUT("/api/themes/:thid", apiHandler(themeHandler(updateTheme)))
	router.DELETE("/api/themes/:thid", apiHandler(themeHandler(deleteTheme)))

	router.GET("/api/themes/:thid/exercices", apiHandler(themeHandler(listThemedExercices)))
	router.POST("/api/themes/:thid/exercices", apiHandler(themeHandler(createExercice)))

	router.GET("/api/themes/:thid/exercices/:eid", apiHandler(exerciceHandler(showExercice)))
	router.PUT("/api/themes/:thid/exercices/:eid", apiHandler(exerciceHandler(updateExercice)))
	router.DELETE("/api/themes/:thid/exercices/:eid", apiHandler(exerciceHandler(deleteExercice)))

	router.GET("/api/themes/:thid/exercices/:eid/files", apiHandler(exerciceHandler(listExerciceFiles)))
	router.POST("/api/themes/:thid/exercices/:eid/files", apiHandler(exerciceHandler(createExerciceFile)))

	router.GET("/api/themes/:thid/exercices/:eid/hints", apiHandler(exerciceHandler(listExerciceHints)))
	router.POST("/api/themes/:thid/exercices/:eid/hints", apiHandler(exerciceHandler(createExerciceHint)))

	router.GET("/api/themes/:thid/exercices/:eid/keys", apiHandler(exerciceHandler(listExerciceFlags)))
	router.POST("/api/themes/:thid/exercices/:eid/keys", apiHandler(exerciceHandler(createExerciceFlag)))

	// Remote
	router.GET("/api/remote/themes", apiHandler(sync.ApiListRemoteThemes))
	router.GET("/api/remote/themes/:thid", apiHandler(sync.ApiGetRemoteTheme))
	router.GET("/api/remote/themes/:thid/exercices", apiHandler(themeHandler(sync.ApiListRemoteExercices)))

	// Synchronize
	router.POST("/api/sync/deep", apiHandler(
		func(_ httprouter.Params, _ []byte) (interface{}, error) {
			return sync.SyncDeep(sync.GlobalImporter), nil
		}))
	router.POST("/api/sync/themes", apiHandler(
		func(_ httprouter.Params, _ []byte) (interface{}, error) {
			return sync.SyncThemes(sync.GlobalImporter), nil
		}))
	router.POST("/api/sync/themes/:thid/exercices", apiHandler(themeHandler(
		func(theme fic.Theme, _ []byte) (interface{}, error) {
			return sync.SyncExercices(sync.GlobalImporter, theme), nil
		})))
	router.POST("/api/sync/themes/:thid/exercices/:eid/files", apiHandler(exerciceHandler(
		func(exercice fic.Exercice, _ []byte) (interface{}, error) {
			return sync.SyncExerciceFiles(sync.GlobalImporter, exercice), nil
		})))
	router.POST("/api/sync/themes/:thid/exercices/:eid/hints", apiHandler(exerciceHandler(
		func(exercice fic.Exercice, _ []byte) (interface{}, error) {
			return sync.SyncExerciceHints(sync.GlobalImporter, exercice), nil
		})))
	router.POST("/api/sync/themes/:thid/exercices/:eid/keys", apiHandler(exerciceHandler(
		func(exercice fic.Exercice, _ []byte) (interface{}, error) {
			return sync.SyncExerciceFlags(sync.GlobalImporter, exercice), nil
		})))

	router.POST("/api/sync/themes/:thid/fixurlid", apiHandler(themeHandler(
		func(theme fic.Theme, _ []byte) (interface{}, error) {
			if theme.FixURLId() {
				return theme.Update()
			}
			return 0, nil
		})))
	router.POST("/api/sync/fixurlids", apiHandler(fixAllURLIds))
}

func fixAllURLIds(_ httprouter.Params, _ []byte) (interface{}, error) {
	nbFix := 0
	if themes, err := fic.GetThemes(); err == nil {
		for _, theme := range themes {
			if theme.FixURLId() {
				theme.Update()
				nbFix += 1
			}

			if exercices, err := theme.GetExercices(); err == nil {
				for _, exercice := range exercices {
					if exercice.FixURLId() {
						exercice.Update()
						nbFix += 1
					}
				}
			}
		}
	}

	return nbFix, nil
}

func bindingFiles(_ httprouter.Params, body []byte) (interface{}, error) {
	if files, err := fic.GetFiles(); err != nil {
		return "", err
	} else {
		ret := ""
		for _, file := range files {
			ret += fmt.Sprintf("%s;%s\n", file.GetOrigin(), file.Path)
		}
		return ret, nil
	}
}

func getExercice(args []string) (fic.Exercice, error) {
	if tid, err := strconv.Atoi(string(args[0])); err != nil {
		return fic.Exercice{}, err
	} else if theme, err := fic.GetTheme(tid); err != nil {
		return fic.Exercice{}, err
	} else if eid, err := strconv.Atoi(string(args[1])); err != nil {
		return fic.Exercice{}, err
	} else {
		return theme.GetExercice(eid)
	}
}

func listThemes(_ httprouter.Params, _ []byte) (interface{}, error) {
	return fic.GetThemes()
}

func exportThemes(_ httprouter.Params, _ []byte) (interface{}, error) {
	return fic.ExportThemes()
}

func showTheme(theme fic.Theme, _ []byte) (interface{}, error) {
	return theme, nil
}

func listThemedExercices(theme fic.Theme, _ []byte) (interface{}, error) {
	return theme.GetExercices()
}

func showThemedExercice(theme fic.Theme, exercice fic.Exercice, body []byte) (interface{}, error) {
	return exercice, nil
}

type uploadedTheme struct {
	Name    string
	URLId   string
	Authors string
	Intro   string
}

func createTheme(_ httprouter.Params, body []byte) (interface{}, error) {
	var ut uploadedTheme
	if err := json.Unmarshal(body, &ut); err != nil {
		return nil, err
	}

	if len(ut.Name) == 0 {
		return nil, errors.New("Theme's name not filled")
	}

	return fic.CreateTheme(ut.Name, ut.URLId, "", ut.Authors, ut.Intro)
}

func updateTheme(theme fic.Theme, body []byte) (interface{}, error) {
	var ut fic.Theme
	if err := json.Unmarshal(body, &ut); err != nil {
		return nil, err
	}

	ut.Id = theme.Id

	if len(ut.Name) == 0 {
		return nil, errors.New("Theme's name not filled")
	}

	if _, err := ut.Update(); err != nil {
		return nil, err
	} else {
		return ut, nil
	}
}

func deleteTheme(theme fic.Theme, _ []byte) (interface{}, error) {
	return theme.Delete()
}
