package sync

import (
	"bufio"
	"crypto"
	"encoding/hex"
	"fmt"
	"io"
	"os"
	"path"
	"strings"

	"srs.epita.fr/fic-server/libfic"

	"gopkg.in/russross/blackfriday.v2"
	_ "golang.org/x/crypto/blake2b"
)

// SyncExerciceHints reads the content of hints/ directories and import it as EHint for the given challenge.
func SyncExerciceHints(i Importer, exercice fic.Exercice) (errs []string) {
	params, err := parseExerciceParams(i, exercice.Path)
	if err != nil {
		errs = append(errs, fmt.Sprintf("%q: challenge.txt: %s", path.Base(exercice.Path), err))
	}

	if _, err := exercice.WipeHints(); err != nil {
		errs = append(errs, err.Error())
	} else {
		for n, hint := range params.Hints {
			if hint.Title != "" {
				hint.Title = fmt.Sprintf("Astuce #%d", n+1)
			}
			if hint.Cost <= 0 {
				hint.Cost = exercice.Gain / 4
			}

			if hint.Filename != "" {
				if hint.Content != "" {
					errs = append(errs, fmt.Sprintf("%q: challenge.txt: hint %s (%d): content and filename can't be filled at the same time", path.Base(exercice.Path), hint.Title, n+1))
					continue
				} else if !i.exists(path.Join(exercice.Path, "hints", hint.Filename)) {
					errs = append(errs, fmt.Sprintf("%q: challenge.txt: hint %s (%d): %s: File not found", path.Base(exercice.Path), hint.Title, n+1, hint.Filename))
					continue
				} else {
					// Handle files as downloadable content
					if res, err := i.importFile(path.Join(exercice.Path, "hints", hint.Filename), func(filePath string, origin string) (interface{}, error) {
						// Calculate hash
						hash512 := crypto.BLAKE2b_512.New()
						if fd, err := os.Open(filePath); err != nil {
							return nil, err
						} else {
							defer fd.Close()

							reader := bufio.NewReader(fd)
							if _, err := io.Copy(hash512, reader); err != nil {
								return nil, err
							}
						}
						result512 := hash512.Sum(nil)

						// Special format for downloadable hints: $FILES + hexhash + path from FILES/
						return "$FILES" + hex.EncodeToString(result512) + strings.TrimPrefix(filePath, fic.FilesDir), nil
					}); err != nil {
						errs = append(errs, fmt.Sprintf("%q: unable to import hint file %q: %s", path.Base(exercice.Path), hint.Filename, err))
						continue
					} else if s, ok := res.(string); !ok {
						errs = append(errs, fmt.Sprintf("%q: unable to import hint file %q: invalid string returned as filename", path.Base(exercice.Path), hint.Filename))
						continue
					} else {
						hint.Content = s
					}
				}
			} else if hint.Content == "" {
				errs = append(errs, fmt.Sprintf("%q: challenge.txt: hint %s (%d): content and filename can't be empty at the same time", path.Base(exercice.Path), hint.Title, n+1))
				continue
			} else {
				hint.Content = string(blackfriday.Run([]byte(hint.Content)))
			}

			// Import hint
			if _, err := exercice.AddHint(hint.Title, hint.Content, hint.Cost); err != nil {
				errs = append(errs, fmt.Sprintf("%q: challenge.txt: hint %s (%d): %s", path.Base(exercice.Path), hint.Title, n+1, err))
			}
		}
	}
	return
}
